/*******************************************************************
*
*  DESCRIPTION: class UControl 
*
*  AUTHOR: Herrero Gustavo
*
*  EMAIL: mailto://gh4o@dc.uba.ar
*
*  DATE: 19/04/2002
*
*******************************************************************/

/** include files **/
#include "ucontrol.h"   // base header
#include "message.h"    // class InternalMessage 
#include "parsimu.h"    // class ParallelMainSimulator
#include "deftp1.h"


/*******************************************************************
* Function Name: UControl
* Description: constructor
********************************************************************/
UControl::UControl( const string &name )
    : Atomic( name )
    , boton_piso( addInputPort( "boton_piso" ) )
    , boton_puerta( addInputPort( "boton_puerta" ) )
    , est_puerta( addInputPort( "est_puerta" ) )
    , resp_motor( addInputPort( "resp_motor" ) )
    , nro_piso_actual( addOutputPort( "nro_piso_actual" ) )
    , sentido_elevador( addOutputPort( "sentido_elevador" ) )
    , subir_cab( addOutputPort( "subir_cab" ) )
    , bajar_cab( addOutputPort( "bajar_cab" ) )
    , abrir_puerta( addOutputPort( "abrir_puerta" ) )
    , cerrar_puerta( addOutputPort( "cerrar_puerta" ) )
{

	if( ParallelMainSimulator::Instance().existsParameter( description(), "piso_inicial" ) )
		piso_inicial = str2Int( ParallelMainSimulator::Instance().getParameter( description(), "piso_inicial" ) );
	else
		piso_inicial = 0;

	if( ParallelMainSimulator::Instance().existsParameter( description(), "cant_pisos" ) )
		cant_pisos = str2Int( ParallelMainSimulator::Instance().getParameter( description(), "cant_pisos" ) );
	else
		cant_pisos = 4;

	if (cant_pisos <= 0)
		piso_inicial = 4;

	if (piso_inicial < 0 || piso_inicial > cant_pisos)
		piso_inicial = 0;

}
    
/*******************************************************************
* Function Name: initFunction
********************************************************************/
Model &UControl::initFunction()
{
	estado_puerta   = s_cerrada;
	estado_motor    = m_detenido;
	piso_actual     = piso_inicial;
	sentido_elev    = e_sin_sentido;
	
	lstSubir.erase( lstSubir.begin(), lstSubir.end());
	lstBajar.erase( lstBajar.begin(), lstBajar.end());
	
	e_proxima_accion  = e_pasivar;
	this->passivate();
	return *this ;
}


/*******************************************************************
* Function Name: externalFunction
* Description: destribuye el evento a la tarea correspondiente.
********************************************************************/
Model &UControl::externalFunction( const ExternalMessage &msg )
{

	if( msg.port() == resp_motor) {
		estado_motor = (int) msg.value();
		switch(estado_motor) {
            case m_detenido:
                 if (sentido_elev == e_subiendo) {
                    //Eliminar el pedido de la lista de pisos a detenerse en subida
            		lstSubir.erase( piso_actual ) ;
                    // determinar el sentido del ascensor
                    if (lstSubir.empty()) {
                        //cambiar de sentido
                        if (!lstBajar.empty())
                            sentido_elev = e_bajando;
                        else
                            sentido_elev = e_sin_sentido;
                    }
                } else{
                    //Eliminar el pedido de la lista de pisos a detenerse en bajada
            		lstBajar.erase( piso_actual ) ;
                    // determinar el sentido del ascensor
                    if (lstBajar.empty()) {
                        //cambiar de sentido
                        if (!lstSubir.empty())
                            sentido_elev = e_subiendo;
                        else
                            sentido_elev = e_sin_sentido;
                    }
                }

        	    // mandar a abrir la puertas
                e_proxima_accion  = e_abrir_puerta;
                holdIn( AtomicState::active, VTime::Zero );
        		break;
        		
            case m_subiendo:
            case m_bajando:
                //solamente actualizar el nro de piso actual y mandar a informarlo.
                piso_actual = piso_actual + estado_motor;
                // mandar a informar el cambio de piso
                e_proxima_accion  = e_informar_cambio_piso;
                holdIn( AtomicState::active, VTime::Zero );
                break;
        }
    }
    


	if( msg.port() == est_puerta) {
	    //solamente reaccionar si el elevador esta detenido
	    if (estado_motor == m_detenido) {
    	    estado_puerta = (int) msg.value();
            switch(estado_puerta) {    
                case s_cerrada:
                    //si se cerraron, programar un ascenso o descenso del elevador.
                    e_proxima_accion  = e_enviar_pedido;
                    holdIn( AtomicState::active, VTime::Zero );
                    break;
            }
        }
    }

	if( msg.port() == boton_puerta) {
	    //solamente reaccionar si el elevador esta detenido
	    if (estado_motor == m_detenido) {
    	    if (msg.value() == bp_cerrar) {
                //mandar a cerrar la puerta.
                e_proxima_accion  = e_cerrar_puerta;
                holdIn( AtomicState::active, VTime::Zero );
            }
    	    if (msg.value() == bp_abrir) {
                //si se cerraron, programar un ascenso o descenso del elevador.
                e_proxima_accion  = e_abrir_puerta;
                holdIn( AtomicState::active, VTime::Zero );
            }

        }
    }

	if( msg.port() == boton_piso && msg.value() >= 0 && msg.value() <= cant_pisos) {
	    
        if (msg.value() == piso_actual && estado_motor == m_detenido) {
            //si se presiona el boton del piso actual, solo abrir la puerta.
            e_proxima_accion  = e_abrir_puerta;
            holdIn( AtomicState::active, VTime::Zero );
            
        } else {
            // encolar el pedido de detencion
    	    if (msg.value() > piso_actual)
        	     lstSubir[(int)msg.value()] = (int) msg.value() ;
            if (msg.value() <= piso_actual)
        	     lstBajar[(int)msg.value()] = (int) msg.value() ;
            
           // determinar el sentido del ascensor
           
            switch (sentido_elev) {
                case e_subiendo:
                    if (lstSubir.empty()) 
                        //cambiar de sentido
                        sentido_elev = e_bajando;
                    break;
                case e_bajando:
                    if (lstBajar.empty())
                        //cambiar de sentido
                        sentido_elev = e_subiendo;
                default: // e_sin_sentido:
                    if (!lstSubir.empty()) 
                        sentido_elev = e_subiendo;
                    else
                        if (!lstBajar.empty()) 
                            sentido_elev = e_bajando;
                    break;
            }
            
            if (estado_puerta == s_cerrada)
                e_proxima_accion  = e_enviar_pedido;
            else
                e_proxima_accion  = e_cerrar_puerta;

            holdIn( AtomicState::active, VTime::Zero );
        }
	}


	return *this;
}

/*******************************************************************
* Function Name: internalFunction
********************************************************************/
Model &UControl::internalFunction( const InternalMessage & )
{
    e_proxima_accion  = e_pasivar;
    passivate();
	return *this ;
}

/*******************************************************************
* Function Name: outputFunction
********************************************************************/
Model &UControl::outputFunction( const CollectMessage &msg )
{
	int out_value;
	
    switch(e_proxima_accion) {
        //mensajes para la puerta
        case e_abrir_puerta:
        	out_value = e_abrir_puerta;
            sendOutput( msg.time(),abrir_puerta,out_value);
            //tambien inforar el sentido del elevador.
            out_value = sentido_elev;
            sendOutput( msg.time(), sentido_elevador, out_value);
            break;
        case e_cerrar_puerta:
        	out_value = e_cerrar_puerta;
            sendOutput( msg.time(),cerrar_puerta,out_value);
            out_value = sentido_elev;
            sendOutput( msg.time(), sentido_elevador, out_value);
            break;

        case e_informar_cambio_piso:
            //mensajes para el mundo exterior
             out_value = piso_actual - piso_inicial;
             sendOutput( msg.time(), nro_piso_actual, out_value);
             out_value = sentido_elev;
             sendOutput( msg.time(), sentido_elevador, out_value);
            break;
        case e_enviar_pedido:
           if (sentido_elev == e_sin_sentido) {
                // dirigirse al piso_inicial
                if (piso_actual < piso_inicial) {
                    lstSubir[piso_inicial] = piso_inicial;
                    sentido_elev = e_subiendo;
                }
                if (piso_actual > piso_inicial) {
                    lstBajar[piso_inicial] = piso_inicial;
                    sentido_elev = e_bajando;
                }
            }
            if (sentido_elev != e_sin_sentido) {
                // deacuerdo al sentido del ascensor mandar el minimo o el maximo 
                //de la lista que corresponda.
                
                int pos_proximo_pisos = piso_actual;

                if (sentido_elev == e_subiendo) {
                    ListaPisos::iterator cursor;
                    
                    /*Buscar el minimo del */
                    for (pos_proximo_pisos = piso_actual;pos_proximo_pisos < cant_pisos; ++pos_proximo_pisos) {
                		//voy buscando el minimo
                		cursor = lstSubir.find(pos_proximo_pisos);
                		if( cursor != lstSubir.end() )
                		    break;
                    }
                    
                     //mensajes para el motor
                     out_value = pos_proximo_pisos - piso_actual;
                     sendOutput( msg.time(), subir_cab, out_value);
                     estado_motor = m_subiendo;
                }
                else {
    
                    ListaPisos::iterator cursor;
    
                    for (pos_proximo_pisos = piso_actual;pos_proximo_pisos >= 0;--pos_proximo_pisos) {
						cursor = lstBajar.find(pos_proximo_pisos);
                		if( cursor != lstBajar.end() )
                		    break;
                        }
                         //mensajes para el motor
                     out_value = piso_actual - pos_proximo_pisos;
                     sendOutput( msg.time(), bajar_cab, out_value);
                     estado_motor = m_bajando;
                }
            }
            out_value = sentido_elev;
            sendOutput( msg.time(), sentido_elevador, out_value);
            break;
   }        

	return *this ;
}

